import numpy as np
from ntsim.photonTransporters.Photons import Photons
from ntsim.utils.report_timing import report_timing
from ntsim.utils.gen_utils import generate_cherenkov_spectrum


import logging
log = logging.getLogger('cascadeParam')

class cascadeCherenkov:
    def __init__(self):
        self.ph_fraction = None
        self.par_tmax_a = 2.00
        self.par_tmax_b = 0.98
        self.par_q_const = 4.13
        log.info("initialized")

    def configure(self, opts):
        self.ph_fraction = 1./opts.photon_suppression
        self.X0 = opts.casc_param_X0
        log.info("configured")

    def rotate_vectors(self,dir,dir0):
        from scipy.spatial.transform import Rotation as R
        dir0 = np.array(dir0)
        phi = np.arctan2(dir0[1],dir0[0])
        theta = np.arccos(dir0[2]/np.sqrt(np.sum(dir0*dir0)))
        phi_d = np.tile([0,0,phi],(len(dir[:,0]),1))
        theta_d = np.tile([0,theta,0],(len(dir[:,0]),1))
        t = R.from_rotvec(phi_d)
        t1 = R.from_rotvec(theta_d)
        dir = t.apply(t1.apply(dir))
        return dir
    
    def sample_positions_and_times(self, n_photons, casc_E, casc_pos, casc_time, casc_dir):
        #
        q_central = self.par_q_const
        tmax_central = self.par_tmax_a + self.par_tmax_b * np.log(casc_E)
        self.q = q_central # TODO: fluctuate it!
        self.tmax = tmax_central # TODO: fluctuate it!
        log.debug(f" | q = {self.q}")
        log.debug(f" | tmax = {self.tmax}")
        #
        x = np.full(shape=n_photons, fill_value=0, dtype=np.float64)
        y = np.full(shape=n_photons, fill_value=0, dtype=np.float64)
        z = np.full(shape=n_photons, fill_value=0, dtype=np.float64)
        #
        gamma_k = self.q + 1
        gamma_theta = self.tmax / (self.q + 1)
        t = np.random.gamma(shape=gamma_k, scale=gamma_theta, size=n_photons)
        dz = t*self.X0
        time = np.full(shape=n_photons, fill_value=casc_time, dtype=np.float64)
        с_const = 0.3 # m / ns
        time += dz / с_const
        z += dz
        position = np.array([x,y,z]).T #[[x1,x2,x3....],[y1,y2,y3....],[z1,z2,z3....]]
        translate_position = np.array(casc_pos)
        position = self.rotate_vectors(position,casc_dir)
        position += translate_position
        return position, time

    def sample_directions(self, n_photons, casc_dir):
        n = 1.35  # TODO Take it from BaikalWater module
        phi = np.random.uniform(low=-np.pi, high=np.pi, size=n_photons)
        costh = 1./n # TODO sample it from distribution!
        sinth = (1.-costh**2)**0.5
        dirx = np.full(shape=n_photons, fill_value=sinth*np.cos(phi), dtype=np.float64)
        diry = np.full(shape=n_photons, fill_value=sinth*np.sin(phi), dtype=np.float64)
        dirz = np.full(shape=n_photons, fill_value=costh, dtype=np.float64)
        direction = np.array([dirx, diry, dirz]).T
        return direction

    def sample_wavelengths(self, n_photons):
        return generate_cherenkov_spectrum(350, 650, n_photons)

    @report_timing
    def propagate(self, vertices):
        log.info("running")
        #
        photons = Photons()
        for vertex in vertices:
            # parse inputs
            pos_time = vertex[0]
            out_particle = vertex[2][0]
            casc_pos  = np.array(pos_time[:3])
            casc_time = vertex[0][3]
            casc_dir  = np.array(out_particle[1:4])
            casc_E    = out_particle[4]
            casc_dir = casc_dir / np.sum(casc_dir**2)**0.5
            log.debug(f" | position (m) : {casc_pos}")
            log.debug(f" | time (ns)    : {casc_time:.3f} ns")
            log.debug(f" | direction    : {casc_dir}")
            log.debug(f" | energy (GeV) : {casc_E:.3f} GeV")
            #
            norm = 1.18e5*casc_E*self.ph_fraction
            n_photons = np.random.poisson(norm)
            log.debug(f" | N = {n_photons}")
            # TODO: fluctuate it!
            # N_mean = f(E)
            # N_std = ?
            # n_photons = np.random.normal(N_mean, N_std)
            # n_photons = int(n_photons)
            ph_pos, ph_time = self.sample_positions_and_times(n_photons, casc_E, casc_pos, casc_time, casc_dir)
            ph_dir = self.sample_directions(n_photons, casc_dir)
            ph_dir = self.rotate_vectors(ph_dir,casc_dir)
            ph_wavelength = self.sample_wavelengths(n_photons)
            #
            # add an extra dummy dimension for further steps
            ph_pos = np.expand_dims(ph_pos, axis=0)
            ph_dir = np.expand_dims(ph_dir, axis=0)
            ph_time = np.expand_dims(ph_time, axis=0)
            #
            photon_bunch = Photons()
            n_steps = 1 # or ph_pos.shape[0]
            photon_bunch.init(n_photons, n_steps, ph_pos, ph_time, ph_dir, ph_wavelength)
            photons = photons.add_photons(photon_bunch)
            
        return photons
