import numpy as np
from ntsim.photonTransporters.Photons import Photons
from ntsim.utils.gen_utils import generate_cherenkov_spectrum
from ntsim.utils.report_timing import report_timing
#from ntsim.utils.gen_utils import rotate_vectors

import logging
log = logging.getLogger('Cerenkov')


class trackCherenkov():
    def __init__(self):
        self.ph_per_m = None
        self.lambda_min = 350.
        self.lambda_max = 650.
        self.ph_fraction = None
        log.info("initialized")

    def configure(self, opts):
        self.ph_fraction = 1./opts.photon_suppression
        self.ph_per_m = opts.track_cherenkov_photons_per_cm * 100.
        log.info("configured")
   
    def rotate_vectors(self,dir,dir0):
        from scipy.spatial.transform import Rotation as R
        dir0 = np.array(dir0)
        phi = np.arctan2(dir0[1],dir0[0])
        theta = np.arccos(dir0[2]/np.sqrt(np.sum(dir0*dir0)))
        phi_d = np.tile([0,0,phi],(len(dir[:,0]),1))
        theta_d = np.tile([0,theta,0],(len(dir[:,0]),1))
        t = R.from_rotvec(phi_d)
        t1 = R.from_rotvec(theta_d)
        #r = t.apply(t1.apply(r))
        dir = t.apply(t1.apply(dir))
        return dir

    def sample_positions_and_times(self, n_photons, r1, r2, t1, t2):
        t = np.random.uniform(t1, t2, size=n_photons)
        r = np.expand_dims(r1, axis=1)
        dr = np.expand_dims(r2-r1, axis=1)
        r = r + dr * (t-t1) / (t2-t1)
        return r.T, t

    def sample_directions(self, n_photons):
        n = 1.35  # TODO Take it from BaikalWater module
        phi = np.random.uniform(low=-np.pi, high=np.pi, size=n_photons)
        costh = 1./n
        sinth = (1.-costh**2)**0.5
        dirx = np.full(shape=n_photons, fill_value=sinth*np.cos(phi), dtype=np.float64)
        diry = np.full(shape=n_photons, fill_value=sinth*np.sin(phi), dtype=np.float64)
        dirz = np.full(shape=n_photons, fill_value=costh, dtype=np.float64)
        direction = np.array([dirx, diry, dirz]).T
        # TODO transform to absolute direction
        
        return direction

    def sample_wavelengths(self, n_photons):
        return generate_cherenkov_spectrum(self.lambda_min, self.lambda_max, n_photons)

    @report_timing
    def propagate(self, tracks):
        log.info("running")
        photons = Photons()
        for key in tracks.keys():
            track = tracks[key]
            pdgid = track.pdg_id
            if pdgid not in [11, -11, 13, -13]:
                continue
            points = np.array(track.points) # [[x,y,z,t,Px,Py,Pz,E], ...]
            R1 = points[:-1,0:3] # segment start positions
            R2 = points[1:,0:3]  # segment end positions
            dR = np.sum((R2 - R1)**2, axis=1)**0.5 # segment lengths
            T1 = points[:-1,3]
            T2 = points[1:,3]
            dT = T2 - T1
            E = points[:-1,7] # segment initial energies
            N_mean = self.ph_per_m * dR * self.ph_fraction
            N = np.random.poisson(N_mean, size=len(dR))
            for r1, r2, dr, t1, t2, n_photons in zip(R1, R2, dR, T1, T2, N):
                if n_photons == 0: continue
                ph_pos, ph_time = self.sample_positions_and_times(n_photons, r1, r2, t1, t2)
                ph_dir = self.sample_directions(n_photons)                
                ph_dir = self.rotate_vectors(ph_dir,r2-r1)
                ph_wavelength = self.sample_wavelengths(n_photons)
                photon_bunch = Photons()
                n_steps = 1
                # add an extra dummy dimension for further steps
                ph_pos = np.expand_dims(ph_pos, axis=0)
                ph_dir = np.expand_dims(ph_dir, axis=0)
                ph_time = np.expand_dims(ph_time, axis=0)
                #
                photon_bunch.init(n_photons, n_steps, ph_pos, ph_time, ph_dir, ph_wavelength)
                photons = photons.add_photons(photon_bunch)
        return photons
